from __future__ import annotations

import sys
from typing import Any, Mapping

from pdm.backend import build_editable as build_editable
from pdm.backend import build_sdist as build_sdist
from pdm.backend import build_wheel as build_wheel
from pdm.backend import (
    prepare_metadata_for_build_editable as prepare_metadata_for_build_editable,
)
from pdm.backend import (
    prepare_metadata_for_build_wheel as prepare_metadata_for_build_wheel,
)

if sys.version_info >= (3, 11):
    import tomllib
else:
    import pdm.backend._vendor.tomli as tomllib


def get_requires_for_build_wheel(
    config_settings: Mapping[str, Any] | None = None,
) -> list[str]:
    with open("pyproject.toml", "rb") as fp:
        config = tomllib.load(fp)
    return config["project"].get("dependencies", [])


get_requires_for_build_sdist = get_requires_for_build_wheel


def get_requires_for_build_editable(
    config_settings: Mapping[str, Any] | None = None,
) -> list[str]:
    return get_requires_for_build_wheel(config_settings) + ["editables"]



from __future__ import annotations

import os
import tarfile
from copy import copy
from io import BytesIO
from pathlib import Path
from posixpath import join as pjoin
from typing import Iterable

from pdm.backend._vendor.packaging.utils import canonicalize_name
from pdm.backend.base import Builder
from pdm.backend.hooks import Context
from pdm.backend.utils import normalize_file_permissions, safe_version, to_filename


def clean_tarinfo(tar_info: tarfile.TarInfo) -> tarfile.TarInfo:
    """
    Clean metadata from a TarInfo object to make it more reproducible.

        - Set uid & gid to 0
        - Set uname and gname to ""
        - Normalise permissions to 644 or 755
        - Set mtime if not None
    """
    ti = copy(tar_info)
    ti.uid = 0
    ti.gid = 0
    ti.uname = ""
    ti.gname = ""
    ti.mode = normalize_file_permissions(ti.mode)

    if "SOURCE_DATE_EPOCH" in os.environ:
        ti.mtime = int(os.environ["SOURCE_DATE_EPOCH"])

    return ti


class SdistBuilder(Builder):
    """This build should be performed for PDM project only."""

    target = "sdist"

    def get_files(self, context: Context) -> Iterable[tuple[str, Path]]:
        collected = dict(super().get_files(context))
        context.ensure_build_dir()
        context.config.write_to(context.build_dir / "pyproject.toml")
        collected["pyproject.toml"] = context.build_dir / "pyproject.toml"
        metadata = self.config.validate()

        def gen_additional_files() -> Iterable[str]:
            if local_hook := self.config.build_config.custom_hook:
                yield local_hook
            if metadata.readme and metadata.readme.file:
                yield metadata.readme.file.relative_to(self.location).as_posix()
            yield from self.find_license_files(metadata)

        root = self.location
        for file in gen_additional_files():
            if file in collected:
                continue
            if root.joinpath(file).exists():
                collected[file] = root / file
        return collected.items()

    def build_artifact(
        self, context: Context, files: Iterable[tuple[str, Path]]
    ) -> Path:
        version = to_filename(safe_version(context.config.metadata["version"]))
        name = to_filename(canonicalize_name(context.config.metadata["name"]))
        dist_info = f"{name}-{version}"

        target = context.dist_dir / f"{dist_info}.tar.gz"

        with tarfile.open(target, mode="w:gz", format=tarfile.PAX_FORMAT) as tar:
            for relpath, path in files:
                tar_info = tar.gettarinfo(path, pjoin(dist_info, relpath))
                tar_info = clean_tarinfo(tar_info)
                if tar_info.isreg():
                    with path.open("rb") as f:
                        tar.addfile(tar_info, f)
                else:
                    tar.addfile(tar_info)
                self._show_add_file(relpath, path)

            pkg_info = str(self.config.validate().as_rfc822()).encode("utf-8")
            tar_info = tarfile.TarInfo(pjoin(dist_info, "PKG-INFO"))
            tar_info.size = len(pkg_info)
            tar_info = clean_tarinfo(tar_info)
            tar.addfile(tar_info, BytesIO(pkg_info))
            self._show_add_file("PKG-INFO", Path("PKG-INFO"))

        return target



# SPDX-License-Identifier: MIT

"""
Constants for the pyproject_metadata package, collected here to make them easy
to update. These should be considered mostly private.
"""

from __future__ import annotations

__all__ = [
    "KNOWN_BUILD_SYSTEM_FIELDS",
    "KNOWN_METADATA_FIELDS",
    "KNOWN_METADATA_VERSIONS",
    "KNOWN_METADATA_VERSIONS",
    "KNOWN_MULTIUSE",
    "KNOWN_PROJECT_FIELDS",
    "KNOWN_TOPLEVEL_FIELDS",
    "PRE_SPDX_METADATA_VERSIONS",
    "PROJECT_TO_METADATA",
]


def __dir__() -> list[str]:
    return __all__


KNOWN_METADATA_VERSIONS = {"2.1", "2.2", "2.3", "2.4"}
PRE_SPDX_METADATA_VERSIONS = {"2.1", "2.2", "2.3"}

PROJECT_TO_METADATA = {
    "authors": frozenset(["Author", "Author-Email"]),
    "classifiers": frozenset(["Classifier"]),
    "dependencies": frozenset(["Requires-Dist"]),
    "description": frozenset(["Summary"]),
    "dynamic": frozenset(),
    "entry-points": frozenset(),
    "gui-scripts": frozenset(),
    "keywords": frozenset(["Keywords"]),
    "license": frozenset(["License", "License-Expression"]),
    "license-files": frozenset(["License-File"]),
    "maintainers": frozenset(["Maintainer", "Maintainer-Email"]),
    "name": frozenset(["Name"]),
    "optional-dependencies": frozenset(["Provides-Extra", "Requires-Dist"]),
    "readme": frozenset(["Description", "Description-Content-Type"]),
    "requires-python": frozenset(["Requires-Python"]),
    "scripts": frozenset(),
    "urls": frozenset(["Project-URL"]),
    "version": frozenset(["Version"]),
}

KNOWN_TOPLEVEL_FIELDS = {"build-system", "project", "tool", "dependency-groups"}
KNOWN_BUILD_SYSTEM_FIELDS = {"backend-path", "build-backend", "requires"}
KNOWN_PROJECT_FIELDS = set(PROJECT_TO_METADATA)

KNOWN_METADATA_FIELDS = {
    "author",
    "author-email",
    "classifier",
    "description",
    "description-content-type",
    "download-url",  # Not specified via pyproject standards, deprecated by PEP 753
    "dynamic",  # Can't be in dynamic
    "home-page",  # Not specified via pyproject standards, deprecated by PEP 753
    "keywords",
    "license",
    "license-expression",
    "license-file",
    "maintainer",
    "maintainer-email",
    "metadata-version",
    "name",  # Can't be in dynamic
    "obsoletes",  # Deprecated
    "obsoletes-dist",  # Rarely used
    "platform",  # Not specified via pyproject standards
    "project-url",
    "provides",  # Deprecated
    "provides-dist",  # Rarely used
    "provides-extra",
    "requires",  # Deprecated
    "requires-dist",
    "requires-external",  # Not specified via pyproject standards
    "requires-python",
    "summary",
    "supported-platform",  # Not specified via pyproject standards
    "version",  # Can't be in dynamic
}

KNOWN_MULTIUSE = {
    "dynamic",
    "platform",
    "provides-extra",
    "supported-platform",
    "license-file",
    "classifier",
    "requires-dist",
    "requires-external",
    "project-url",
    "provides-dist",
    "obsoletes-dist",
    "requires",  # Deprecated
    "obsoletes",  # Deprecated
    "provides",  # Deprecated
}



from __future__ import annotations

import dataclasses
from pathlib import Path
from typing import TYPE_CHECKING, Any, Iterable

from pdm.backend.config import Config

if TYPE_CHECKING:
    from typing import Protocol

    from pdm.backend.base import Builder, Target
else:
    Protocol = object


@dataclasses.dataclass()
class Context:
    """The context object for the build hook,
    which contains useful information about the building process.
    Custom hooks can also change the values of attributes or
    assign arbitrary attributes to this object.

    Attributes:
        build_dir: The build directory for storing files generated during the build
        dist_dir: The directory to store the built artifacts
        kwargs: The extra args passed to the build method
        builder: The builder associated with this build context
    """

    build_dir: Path
    dist_dir: Path
    kwargs: dict[str, Any]
    builder: Builder

    @property
    def config(self) -> Config:
        """The parsed pyproject.toml as a Config object"""
        return self.builder.config

    @property
    def root(self) -> Path:
        """The project root directory"""
        return self.builder.location

    @property
    def target(self) -> Target:
        """The target to build, one of 'sdist', 'wheel', 'editable'"""
        return self.builder.target

    @property
    def config_settings(self) -> dict[str, str]:
        """The config settings passed to the hook"""
        return self.builder.config_settings

    def ensure_build_dir(self) -> Path:
        """Return the build dir and create if it doesn't exist"""
        if not self.build_dir.exists():
            self.build_dir.mkdir(mode=0o700, parents=True, exist_ok=True)
            (self.build_dir / ".gitignore").write_text("*\n")
        return self.build_dir

    def expand_paths(self, path: str) -> Iterable[Path]:
        def path_filter(p: Path) -> bool:
            return p.is_file() or p.is_symlink()

        plib_path = Path(path)
        if plib_path.parts and plib_path.parts[0] == "${BUILD_DIR}":
            return filter(
                path_filter, self.build_dir.glob(Path(*plib_path.parts[1:]).as_posix())
            )

        return filter(path_filter, self.root.glob(path))


class BuildHookInterface(Protocol):
    """The interface definition for build hooks.

    Custom hooks can implement part of the methods to provide corresponding abilities.
    """

    def pdm_build_hook_enabled(self, context: Context) -> bool:
        """Return True if the hook is enabled for the current build and context

        Parameters:
            context: The context for this build
        """
        ...

    def pdm_build_clean(self, context: Context) -> None:
        """An optional clean step which will be called before the build starts

        Parameters:
            context: The context for this build
        """
        ...

    def pdm_build_initialize(self, context: Context) -> None:
        """This hook will be called before the build starts,
        any updates to the context object will be seen by the following processes.
        It is recommended to modify the metadata in this hook.

        Parameters:
            context: The context for this build
        """
        ...

    def pdm_build_update_files(self, context: Context, files: dict[str, Path]) -> None:
        """Passed in the current file mapping of {relpath: path}
        for hooks to update.

        Parameters:
            context: The context for this build
            files: The file mapping to be included in the build artifact, where the key
                is the relpath inside the artifact(wheel or tarball) and the value is
                the local path to the file.
        """
        ...

    def pdm_build_finalize(self, context: Context, artifact: Path) -> None:
        """This hook will be called after the build is done,
        the artifact is the path to the built artifact.

        Parameters:
            context: The context for this build
            artifact: The path to the built artifact
        """
        ...

    def pdm_build_update_setup_kwargs(
        self, context: Context, kwargs: dict[str, Any]
    ) -> None:
        """Passed in the setup kwargs for hooks to update.

        Parameters:
            context: The context for this build
            kwargs: The arguments to be passed to the setup() function

        Note:
            This hook will be called in the subprocess of running setup.py.
            Any changes made to the context won't be written back.
        """
        ...
